/****************************************************************************
* VCGLib                                                            o o     *
* Visual and Computer Graphics Library                            o     o   *
*                                                                _   O  _   *
* Copyright(C) 2004-2012                                           \/)\/    *
* Visual Computing Lab                                            /\/|      *
* ISTI - Italian National Research Council                           |      *
*                                                                    \      *
* All rights reserved.                                                      *
*                                                                           *
* This program is free software; you can redistribute it and/or modify      *
* it under the terms of the GNU General Public License as published by      *
* the Free Software Foundation; either version 2 of the License, or         *
* (at your option) any later version.                                       *
*                                                                           *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
* GNU General Public License (http://www.gnu.org/licenses/gpl.txt)          *
* for more details.                                                         *
*                                                                           *
****************************************************************************/
/*! \file trimesh_base.cpp
\ingroup code_sample

\brief the minimal example of using the lib

This file contain a minimal example of the library, showing how to load a mesh and how to compute per vertex normals on it.

*/

#include<vcg/complex/complex.h>
#include<wrap/io_trimesh/import_off.h>
#include<wrap/io_trimesh/export_off.h>
#include<wrap/io_trimesh/export_obj.h>
#include<wrap/io_trimesh/export_ply.h>
#include<vcg/complex/algorithms/update/topology.h>
#include<vcg/complex/algorithms/update/normal.h>
#include<vcg/complex/algorithms/clean.h>

class MyVertex; class MyEdge; class MyFace;
struct MyUsedTypes : public vcg::UsedTypes<vcg::Use<MyVertex>   ::AsVertexType,
                                           vcg::Use<MyEdge>     ::AsEdgeType,
                                           vcg::Use<MyFace>     ::AsFaceType>{};

class MyVertex  : public vcg::Vertex< MyUsedTypes, vcg::vertex::Coord3f, vcg::vertex::Normal3f, vcg::vertex::BitFlags, vcg::vertex::Color4b, vcg::vertex::VEAdj  >{};
class MyFace    : public vcg::Face<   MyUsedTypes, vcg::face::FFAdj,  vcg::face::VertexRef, vcg::face::Color4b,  vcg::face::BitFlags > {};
class MyEdge    : public vcg::Edge<   MyUsedTypes, vcg::edge::VertexRef, vcg::edge::VEAdj, vcg::edge::BitFlags> {};

class MyMesh    : public vcg::tri::TriMesh< std::vector<MyVertex>, std::vector<MyFace> , std::vector<MyEdge>  > {};
using namespace vcg;



int countNonVisitedEdges(MyVertex * vp, MyEdge * &ep)
{
  std::vector<MyEdge *> starVec;
  edge::VEStarVE(&*vp,starVec);

  int cnt =0;
  for(size_t i=0;i<starVec.size();++i)
  {
    if(!starVec[i]->IsV())
    {
      cnt++;
      ep = starVec[i];
    }
  }

  return cnt;
}

void Retract(MyMesh &t)
{
  tri::Clean<MyMesh>::RemoveDuplicateVertex(t);
  printf("Retracting a mesh of %i edges and %i vertices\n",t.en,t.vn);
  tri::UpdateTopology<MyMesh>::VertexEdge(t);

  std::stack<MyVertex *> vertStack;

  for(MyMesh::VertexIterator vi=t.vert.begin();vi!=t.vert.end();++vi)
  {
    std::vector<MyEdge *> starVec;
    edge::VEStarVE(&*vi,starVec);
    if(starVec.size()==1)
      vertStack.push(&*vi);
  }

  tri::UpdateFlags<MyMesh>::EdgeClearV(t);
  tri::UpdateFlags<MyMesh>::VertexClearV(t);

  while(!vertStack.empty())
  {
    MyVertex *vp = vertStack.top();
    vertStack.pop();
    vp->C()=Color4b::Blue;
    MyEdge *ep=0;

    int eCnt =  countNonVisitedEdges(vp,ep);
    if(eCnt==1) // We have only one non visited edge over vp
    {
      assert(!ep->IsV());
      ep->SetV();
      MyVertex *otherVertP;
      if(ep->V(0)==vp) otherVertP = ep->V(1);
      else otherVertP = ep->V(0);
      vertStack.push(otherVertP);
    }
  }

  tri::io::ExporterPLY<MyMesh>::Save(t,"retract0.ply",tri::io::Mask::IOM_EDGEINDEX | tri::io::Mask::IOM_VERTCOLOR);

  for(size_t i =0; i<t.edge.size();++i)
    if(t.edge[i].IsV()) tri::Allocator<MyMesh>::DeleteEdge(t,t.edge[i]);

  tri::Clean<MyMesh>::RemoveUnreferencedVertex(t);

  tri::Allocator<MyMesh>::CompactEveryVector(t);
  tri::io::ExporterPLY<MyMesh>::Save(t,"retract1.ply",tri::io::Mask::IOM_EDGEINDEX | tri::io::Mask::IOM_VERTCOLOR);

  for(size_t i =0; i<t.edge.size();++i)
    if(t.edge[i].IsV()) tri::Allocator<MyMesh>::DeleteEdge(t,t.edge[i]);

  tri::Allocator<MyMesh>::CompactEveryVector(t);

  for(int k=0;k<50;++k)
  {
    std::vector<Point3f> posVec(t.vn,Point3f(0,0,0));
    std::vector<int> cntVec(t.vn,0);

    for(size_t i =0; i<t.en;++i)
      for(int j=0;j<2;++j)
      {
        int vertInd = tri::Index(t,t.edge[i].V(j));
        posVec[vertInd] += t.edge[i].V1(j)->P();
        cntVec[vertInd] ++;
      }

    for(size_t i =0; i<t.en;++i)
      t.vert[i].P() = (t.vert[i].P() + posVec[i])/float(cntVec[i]+1);
  }
  tri::io::ExporterPLY<MyMesh>::Save(t,"retract2.ply",tri::io::Mask::IOM_EDGEINDEX | tri::io::Mask::IOM_VERTCOLOR);

}

int main( int argc, char **argv )
{
  if(argc<2)
  {
    printf("Usage trimesh_base <meshfilename.off>\n");
    return -1;
  }
  MyMesh m;
  tri::io::ImporterOFF<MyMesh>::Open(m,argv[1]);
  printf( "Mesh has %i vert and %i faces\n", m.VN(), m.FN() );
  tri::UpdateTopology<MyMesh>::FaceFace(m);

  tri::UpdateFlags<MyMesh>::FaceClearV(m);

  std::vector<face::Pos<MyFace> > visitStack;

  typedef face::Pos<MyFace> MyPos;
  MyMesh treeMesh, dualMesh;
  m.face[0].SetV();
  for(int i=0;i<3;++i)
    visitStack.push_back(MyPos(&(m.face[0]),i,m.face[0].V(i)));

  int cnt=1;

  while(!visitStack.empty())
  {
    std::swap(visitStack.back(),visitStack[rand()%visitStack.size()]);
    MyPos c = visitStack.back();
    visitStack.pop_back();
    assert(c.F()->IsV());
    c.F()->C() = Color4b::ColorRamp(0,m.fn,cnt);
    c.FlipF();
    if(!c.F()->IsV())
    {
      tri::Allocator<MyMesh>::AddEdge(treeMesh,Barycenter(*(c.FFlip())),Barycenter(*(c.F())));
      ++cnt;
      c.F()->SetV();
      c.FlipE();c.FlipV();
      visitStack.push_back(c);
      c.FlipE();c.FlipV();
      visitStack.push_back(c);
    }
    else
    {
      tri::Allocator<MyMesh>::AddEdge(dualMesh,c.V()->P(),c.VFlip()->P());
    }
  }
  assert(cnt==m.fn);

  tri::io::ExporterOFF<MyMesh>::Save(m,"out.off",tri::io::Mask::IOM_FACECOLOR);
  tri::io::ExporterPLY<MyMesh>::Save(treeMesh,"tree.ply",tri::io::Mask::IOM_EDGEINDEX);
  tri::io::ExporterPLY<MyMesh>::Save(dualMesh,"dual.ply",tri::io::Mask::IOM_EDGEINDEX);

  Retract(dualMesh);
  return 0;
}









